local FilterWorkplaces = function(unit)
  local dome_workplaces
  local cur_workplace = unit.workplace
  if cur_workplace and unit.lock_workplace then
    if IsKindOf(cur_workplace, "Workplace") then
      dome_workplaces = {cur_workplace}
    end
  elseif unit:CanWork() then
    local dome = unit.dome
    local dome_dist = dome:GetOutsideWorkplacesDist()
    local avoid_workplace = unit.avoid_workplace
    if avoid_workplace and (not unit.avoid_workplace_start or unit.avoid_workplace_start + g_Consts.AvoidWorkplaceSols < unit.city.day) then
      avoid_workplace = false
    end
    local specialist = unit.specialist
    for _, workplace in ipairs(unit.city.labels.Workplace or empty_table) do
      local parent_dome = workplace.parent_dome
      if not workplace.destroyed and not workplace.demolishing and (parent_dome == dome or not parent_dome and dome_dist >= HexAxialDistance(workplace, dome)) and workplace.ui_working and workplace:CanAcceptWorkers() and workplace ~= avoid_workplace and (workplace:GetFreeSlots() > 0 or workplace.specialist == specialist and workplace:GetWorkerToKick(unit, false)) then
        dome_workplaces = dome_workplaces or {}
        dome_workplaces[#dome_workplaces + 1] = workplace
      end
    end
  end
  return dome_workplaces
end

local GetShifts = function(bld, is_training)
  if is_training then
    return #bld.visitors
  end
  return #bld.workers
end

VanillaChooseWorkplace = ChooseWorkplace

function ChooseFilteredWorkplace(unit, work_type)
  local dome_workplaces = FilterWorkplaces(unit)
  if not dome_workplaces then
    return unit.workplace, unit.workplace_shift, false
  end
  local current_bld = unit.workplace
  local current_shift = unit.workplace_shift
  local current_priority = current_bld and current_bld.priority or 0
  local unit_specialization = unit.specialist
  local to_kick, current_workers, max_workers = GetWorkersCount(current_bld, current_shift, unit, current_bld and current_bld:IsKindOf("TrainingBuilding"))
  current_workers = current_workers - to_kick
  local current_workplace_percent = current_bld and max_workers > 0 and MulDivRound(current_workers, 100, max_workers) or 0
  local current_workplace_percent_after = current_bld and max_workers > 0 and MulDivRound(current_workers - 1, 100, max_workers) or 0
  local training_now = current_bld and not current_bld.is_workplace
  local best_bld, best_shift, best_priority, best_kicked
  for i = const.MaxBuildingPriority, 0, -1 do
    if i < current_priority then
      return current_bld, current_shift, false
    end
    if not current_bld or i > current_priority then
      local score
      for _, workplace in ipairs(dome_workplaces) do
        if workplace.priority == i then
          local nShifts = GetShifts(workplace)
          for k = 1, nShifts do
            local kick_candidate
            local to_kick, count_workers, max_workers = GetWorkersCount(workplace, k, unit)
            local current_workers = count_workers - to_kick
            if max_workers > 0 then
              local new_workforce_after = MulDivRound(current_workers + 1, 100, max_workers)
              if (not score or score > new_workforce_after or best_bld.specialist ~= unit_specialization and workplace.specialist == unit_specialization) and (workplace.specialist == unit_specialization or workplace.allow_all_workers) then
                local free_workplace = count_workers < max_workers
                has_space = free_workplace or to_kick > 0
				kick_candidate = workplace:GetWorkerToKick(unit, k)
                if has_space then
                  score = new_workforce_after
                  best_bld = workplace
                  best_shift = k
                  best_priority = i
				  if not free_workplace then
				    best_kicked = kick_candidate
				  else
				    best_kicked = false
				  end
                end
              end
            end
          end
        end
      end
      if score then
        return best_bld, best_shift, best_kicked
      end
    end
    if current_priority == i then
      local score
      for _, workplace in ipairs(dome_workplaces) do
        if workplace.priority == i then
          local nShifts = GetShifts(workplace)
          for k = 1, nShifts do
            local kick_candidate
            local to_kick, count_workers, max_workers = GetWorkersCount(workplace, k, unit)
            local current_workers = count_workers - to_kick
            if max_workers > 0 and (unit.workplace_shift ~= k or workplace ~= current_bld) then
              local new_workforce_before = MulDivRound(current_workers, 100, max_workers)
              local new_workforce_after = MulDivRound(current_workers + 1, 100, max_workers)
              local free_workplace = count_workers < max_workers
              has_space = free_workplace or to_kick > 0
			  kick_candidate = workplace:GetWorkerToKick(unit, k)
              if kick_candidate and (best_bld and best_bld.specialist ~= unit_specialization or current_bld and current_bld.specialist ~= unit_specialization) and workplace.specialist == unit_specialization then
                score = new_workforce_after
                best_bld = workplace
                best_shift = k
                best_priority = i
				best_kicked = kick_candidate
              elseif free_workplace and current_workplace_percent_after > new_workforce_before and (not score or new_workforce_after > score or best_bld.specialist ~= unit_specialization and workplace.specialist == unit_specialization) and (workplace.specialist == unit_specialization or workplace.allow_all_workers) then
                score = new_workforce_after
                best_bld = workplace
                best_shift = k
                best_priority = i
                best_kicked = false
              end
            end
          end
        end
      end
      return best_bld or current_bld, best_shift or current_shift, best_kicked
    end
  end
  return best_bld, best_shift
end

ChooseWorkplace = ChooseFilteredWorkplace